// LPIIR.cpp: implementation of the CLPIIR class.
//    A 3rd order LP complex IIR section.
//////////////////////////////////////////////////////////////////////
// Copyright 2000.    Moe Wheatley AE4JY  <ae4jy@mindspring.com>
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either version 2
//of the License, or any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "LPIIR.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CLPIIR::CLPIIR()
{

}

CLPIIR::~CLPIIR()
{

}

//////////////////////////////////////////////////////////////////////
//  Sets the filter coefficients depending on filter number.  
//
//   B1, B0, A1   for the 1st order section
//
//   B2, B1, B0, A2, A1  for the 2nd order section
//
//////////////////////////////////////////////////////////////////////
void CLPIIR::Init(const double* pLPCoef)
{
	m_IIR1_B1 = pLPCoef[0];
	m_IIR1_B0 = pLPCoef[1];
	m_IIR1_A1 = pLPCoef[2];
//
	m_IIR2_B2 = pLPCoef[3];
	m_IIR2_B1 = pLPCoef[4];
	m_IIR2_B0 = pLPCoef[5];
	m_IIR2_A2 = pLPCoef[6];
	m_IIR2_A1 = pLPCoef[7];
//
	m_W1 = 0.0;
//
	m_W2[0] = 0.0;
	m_W2[1] = 0.0;
}

//////////////////////////////////////////////////////////////////////
//  Calculate complex 3rd order IIR LP filter
//////////////////////////////////////////////////////////////////////
double CLPIIR::CalcFilter(double in)
{
	return IIRFilter2( IIRFilter1(in) );
}


/*==================================================================*/
/*  Function: IIRFilter1                                            */
/*                                                                  */
/*  Implements a first order IIR filter stage using canonical form. */
/*  The transfer function of the filter stage implemented is in     */
/*  the form                                                        */
/*                                                                  */
/*                          -1                                      */
/*                 B0 + B1 z                                        */
/*          H(z) = -----------                                      */
/*                         -1                                       */
/*                  1 + A1 z                                        */
/*                                                                  */
/*  The block diagram used in the implementation is given below:    */
/*                                                                  */
/*          input               w(n)   B0         output            */
/*            -----> + ----------+-----[>--> + ----->               */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A1     |w(n-1) B1  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                                                                  */
/*==================================================================*/
double CLPIIR::IIRFilter1(double in)
{
double w;
double out;
			// w(n) = w(n) - A1*w(n-1)
	w = in - (m_IIR1_A1 * m_W1);	
			// out(n) = B0*w(n) + B1*w(n-1)
	out  = m_IIR1_B0 * w + m_IIR1_B1 * m_W1;	
			//w(n-1) = w(n)
	m_W1 = w;
	return out;
}


/*==================================================================*/
/*  Function: IIRFilter2                                            */
/*                                                                  */
/*  Implements a second order IIR filter stage in canonical form.   */
/*  The transfer function of the filter stage implemented is in     */
/*  the form                                                        */
/*                                                                  */
/*                          -1       -2                             */
/*                 B0 + B1 z   + B2 z                               */
/*          H(z) = --------------------                             */
/*                         -1       -2                              */
/*                 1 + A1 z   + A2 z                                */
/*                                                                  */
/*  The block diagram used in the implementation is given below:    */
/*                                                                  */
/*          input               w(n)   B0         output            */
/*            -----> + ----------+-----[>--> + ----->               */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A1     |w(n-1) B1  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A2     |w(n-2) B2  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                                                                  */
/*==================================================================*/
double CLPIIR::IIRFilter2(double in)
{
double w;
double out;
//		w(n) = in - A1*w(n-1) - A2*w(n-2)
	w  = in - m_IIR2_A1*m_W2[0] - m_IIR2_A2*m_W2[1];
//		out = B0*w(n) + B1*w(n-1) + B2*w(n-2)
	out = m_IIR2_B0*w + m_IIR2_B1*m_W2[0] + m_IIR2_B2*m_W2[1];
//		w(n-2) = w(n-1)       w(n-1) = w(n)
	m_W2[1] = m_W2[0];
	m_W2[0] = w;
	return out;
}
